/*####COPYRIGHTBEGIN####
 -------------------------------------------
 Copyright (C) 2003 Steve Karg

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to:
 The Free Software Foundation, Inc.
 59 Temple Place - Suite 330
 Boston, MA  02111-1307
 USA.

 As a special exception, if other files instantiate templates or
 use macros or inline functions from this file, or you compile
 this file and link it with other works to produce a work based
 on this file, this file does not by itself cause the resulting
 work to be covered by the GNU General Public License. However
 the source code for this file must still be made available in
 accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work
 based on this file might be covered by the GNU General Public
 License.

 Modified by Kåre Särs (kare.sars@wapice.com) March 2005
 -------------------------------------------
####COPYRIGHTEND####*/

/* This clause describes a Master-Slave/Token-Passing (MS/TP) data link  */
/* protocol, which provides the same services to the network layer as  */
/* ISO 8802-2 Logical Link Control. It uses services provided by the  */
/* EIA-485 physical layer. Relevant clauses of EIA-485 are deemed to be  */
/* included in this standard by reference. The following hardware is assumed: */
/* (a)	A UART (Universal Asynchronous Receiver/Transmitter) capable of */
/*      transmitting and receiving eight data bits with one stop bit  */
/*      and no parity. */
/* (b)	An EIA-485 transceiver whose driver may be disabled.  */
/* (c)	A timer with a resolution of five milliseconds or less */

#include "platform.h"
#include "conf.h"
#include "npdu.h"
#include "mstp.h"
#include "mstp_crc.h"
#include "rs485.h"

/* MS/TP Frame Format */
/* All frames are of the following format: */
/* */
/* Preamble: two octet preamble: X`55', X`FF' */
/* Frame Type: one octet */
/* Destination Address: one octet address */
/* Source Address: one octet address */
/* Length: two octets, most significant octet first, of the Data field */
/* Header CRC: one octet */
/* Data: (present only if Length is non-zero) */
/* Data CRC: (present only if Length is non-zero) two octets, */
/*           least significant octet first */
/* (pad): (optional) at most one octet of padding: X'FF' */

  /* The number of tokens received or used before a Poll For Master cycle */
/* is executed: 50. (0-49) */
#define Npoll 48

/* The number of retries on sending Token: 1. */
#define Nretry_token 1

/* The minimum number of DataAvailable or ReceiveError events that must be */
/* seen by a receiving node in order to declare the line "active": 4. */
#define Nmin_octets 4

/* The minimum time without a DataAvailable or ReceiveError event within */
/* a frame before a receiving node may discard the frame: 60 bit times. */
/* (Implementations may use larger values for this timeout, */
/* not to exceed 100 milliseconds.) */
/* At 9600 baud, 60 bit times would be about 6.25 milliseconds */
// defined in mstp_port
//#define Tframe_abort bit_times_to_ms(60, 9600)

/* The maximum idle time a sending node may allow to elapse between octets */
/* of a frame the node is transmitting: 20 bit times. */
/*#define Tframe_gap bit_times_to_ms(20, 9600)*/

/* The time without a DataAvailable or ReceiveError event before declaration */
/* of loss of token: 500 milliseconds. */
#define Tno_token 500

/* The maximum time after the end of the stop bit of the final */
/* octet of a transmitted frame before a node must disable its */
/* EIA-485 driver: 15 bit times. */
/*#define Tpostdrive bit_times_to_ms(15, 9600)*/

/* The maximum time a node may wait after reception of a frame that expects */
/* a reply before sending the first octet of a reply or Reply Postponed */
/* frame: 250 milliseconds. */
#define Treply_delay_max 250
#define Treply_delay 200

/* The minimum time without a DataAvailable or ReceiveError event */
/* that a node must wait for a station to begin replying to a */
/* confirmed request: 255 milliseconds. (Implementations may use */
/* larger values for this timeout, not to exceed 300 milliseconds.) */
#define Treply_timeout 255

/* Repeater turnoff delay. The duration of a continuous logical one state */
/* at the active input port of an MS/TP repeater after which the repeater */
/* will enter the IDLE state: 29 bit times < Troff < 40 bit times. */
/* #define Troff 30 */

/* The width of the time slot within which a node may generate a token: */
/* 10 milliseconds. */
#define Tslot 10

/* The maximum time a node may wait after reception of the token or */
/* a Poll For Master frame before sending the first octet of a frame: */
/* 15 milliseconds. */
#define Tusage_delay 10

/* The minimum time without a DataAvailable or ReceiveError event that a */
/* node must wait for a remote node to begin using a token or replying to */
/* a Poll For Master frame: 20 milliseconds. (Implementations may use */
/* larger values for this timeout, not to exceed 100 milliseconds.) */
//#define Tusage_timeout 20
#define Tusage_timeout 30

/* The minimum time after the end of the stop bit of the final octet of a */
/* received frame before a node may enable its EIA-485 driver: 40 bit times. */
/* At 9600 baud, 40 bit times would be about 4.166 milliseconds */
// defined in mstp_port
//#define Tturnaround bit_times_to_ms(40,9600)


/* auto baud rate parameters */
#define AUTO_BOUD_BAUD_FOUND 100

struct mstp_port_struct_t *mstp_port;

/**********************************************************************/
void MSTP_Master_Node_State_Machine(void);



/**********************************************************************/
INT16U MSTP_Create_Frame(
        INT8U *buffer, /* where frame is loaded */
        INT16U buffer_len, /* amount of space available */
        INT8U frame_type, /* type of frame to send - see defines */
        INT8U destination, /* destination address */
        /* source address (mstp_port->This_Station)*/
        INT8U *data, /* any data to be sent - may be null */
        INT16U data_len) /* number of bytes of data (up to 501) */
{
    INT8U crc8 = 0xFF; /* used to calculate the crc value */
    INT16U crc16 = 0xFFFF; /* used to calculate the crc value */
    INT16U index = 0; /* used to load the data portion of the frame */

    /* not enough to do a header */
    if (buffer_len < 8)
        return 0;

    buffer[0] = 0x55;
    buffer[1] = 0xFF;
    buffer[2] = frame_type;
    crc8 = CRC_Calc_Header(buffer[2],crc8);
    buffer[3] = destination;
    crc8 = CRC_Calc_Header(buffer[3],crc8);
    buffer[4] = mstp_port->This_Station; /* should always be this station */
    crc8 = CRC_Calc_Header(buffer[4],crc8);
    buffer[5] = (INT8U)(data_len / 256);
    crc8 = CRC_Calc_Header(buffer[5],crc8);
    buffer[6] = data_len % 256;
    crc8 = CRC_Calc_Header(buffer[6],crc8);
    buffer[7] = ~crc8;

    index = 8;

    while (data_len && data && (index < buffer_len))
    {
        buffer[index] = *data;
        crc16 = CRC_Calc_Data(buffer[index],crc16);
        data++;
        index++;
        data_len--;
    }
    /* append the data CRC if necessary */
    if (index > 8)
    {
        if ((index + 2) <= buffer_len)
        {
            crc16 = ~crc16;
            buffer[index] = (INT8U)(crc16&0xff);
            index++;
            buffer[index] = (INT8U)((crc16>>8)&0xff);
            index++;
        }
        else
            return 0;
    }

    return index; /* returns the frame length */
}


/* MSTP_Send_Data_Frame is always called from an interrupt function */
/****************************************************************/
BOOLEAN MSTP_Send_Data_Frame(INT8U *obuffer, INT16U olen)
{
    if ((mstp_port->Tturnaround > mstp_port->SilenceTimer) || (mstp_port->TxEnabeled==TRUE)) {
        /* delay sending */
        return FALSE;
    }

    if (mstp_port->receive_state!=MSTP_RECEIVE_STATE_IDLE) {
        one_p_printf("receive_state == %d != MSTP_RECEIVE_STATE_IDLE\n", mstp_port->receive_state);
        mstp_port->receive_state=MSTP_RECEIVE_STATE_IDLE;
    }

    (void)RS485_Send_First_Byte(obuffer, olen);

    return TRUE;
}


/* MSTP_Send_Controll_Frame is always called from an interrupt function */
/************************************************************************/
BOOLEAN MSTP_Send_Controll_Frame(INT8U frame_type, /* type of frame to send - see defines */
                                 INT8U destination)/* destination address */
{

    if ((mstp_port->Tturnaround > mstp_port->SilenceTimer) || (mstp_port->TxEnabeled==TRUE)) {
        /* delay sending */
        return FALSE;
    }

    if (mstp_port->receive_state!=MSTP_RECEIVE_STATE_IDLE) {
        one_p_printf("receive_state == %d != MSTP_RECEIVE_STATE_IDLE\n", mstp_port->receive_state);
        mstp_port->receive_state=MSTP_RECEIVE_STATE_IDLE;
    }

    if (MSTP_Create_Frame(mstp_port->O_Buffer_Controll,
        MSTP_HEADER_SIZE,
        frame_type,
        destination,
        NULL,
        0) == 0)
    {
        no_p_printf("Creating frame FAILED!\n");
    }

    (void)RS485_Send_First_Byte(mstp_port->O_Buffer_Controll, MSTP_HEADER_SIZE);

    return TRUE;
}

/* Millisecond Timer - called every x millisecond(s) */
/****************************************************************/
void MSTP_Millisecond_Timer(INT8U delay)
{

    if ((mstp_port != NULL) && (mstp_port->run_msec_timer == TRUE))
    {
        /* SilenceTimer */
        mstp_port->SilenceTimer += delay;  /* increment timer */
        mstp_port->SilenceTimer &= 0x7FFF; /* if we get over 0x7FFF return to 0 */

        /* ReplyPostponedTimer */
        if (mstp_port->ReplyPostponedTimer < Treply_delay_max+1) mstp_port->ReplyPostponedTimer+=delay;

        /* Run the Master_Node_State_Machine */
        d_no_p_printf("run MSTP_Master_Node_State_Machine\n");
        MSTP_Master_Node_State_Machine();
    }

    return;
}

/****************************************************************/
void MSTP_Receive_Frame_State_Machine(void)
{
    switch (mstp_port->receive_state)
    {
        /* In the IDLE state, the node waits for the beginning of a frame. */
        case MSTP_RECEIVE_STATE_IDLE:
            /* EatAnError */
            if (mstp_port->ReceiveError == TRUE)
            {
                fr_printf("R_IDLE EatAnError\n");
                mstp_port->ReceiveError = FALSE;
                //mstp_port->SilenceTimer = 0;
                mstp_port->EventCount++;
                /* wait for the start of a frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            else if (mstp_port->DataAvailable == TRUE)
            {
                /* Preamble1 */
                if (mstp_port->DataRegister == 0x55)
                {
                    d_fr_printf("R_IDLE Preamble1\n");
                    mstp_port->DataAvailable = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    /* receive the remainder of the frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_PREAMBLE;
                }
                /* EatAnOctet */
                else
                {
                    d_fr_printf("R_IDLE EatAnOctet\n");
                    mstp_port->DataAvailable = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    /* wait for the start of a frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;

                    /* Hack to get auto baud rate */
                    /* if we continuously receice garbage we propably have wrong baud rate, */
                    /* but if we already have found one baudrate changing it must be manual */
                    if (mstp_port->found_frames < AUTO_BOUD_BAUD_FOUND) {
                        mstp_port->looking_for_frame++;
                        if ((mstp_port->looking_for_frame % 100) == 0) {
                            d_one_p_printf("looking_for_frame = %d\n", mstp_port->looking_for_frame);
                        }
                        if (mstp_port->looking_for_frame > (MAX_MSTP_PDU_SIZE * 2)) {
                            /* Worst case is where a node sends max number of octets and has a header crc error */
                            /* If two max sized cells have crc errors we are in trouble */
                            mstp_port->looking_for_frame=0;
                            no_p_printf("try_next_baudrate\n");
                            mstp_port->npdu_r_err |= READ_TRY_NEXT_BR;
                            bac_sem_post(mstp_port->npdu_read_sem);
                        }
                    }
                }
            }
            break;
            /* In the PREAMBLE state, the node waits for the second octet of the preamble. */
        case MSTP_RECEIVE_STATE_PREAMBLE:
            /* Timeout */
            if (mstp_port->SilenceTimer > mstp_port->Tframe_abort)
            {
                fr_printf("R_PREAMBLE Timeout \n");
                /* a correct preamble has not been received */
                /* (ReceiveError and DataAvailable must be false) */
                /* wait for the start of a frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }

            /* Error */
            else if (mstp_port->ReceiveError == TRUE)
            {
                /* DataAvailable must be FALSE */
                fr_printf("R_PREAMBLE Error\n");
                mstp_port->ReceiveError = FALSE;
                //mstp_port->SilenceTimer = 0;
                mstp_port->EventCount++;
                /* wait for the start of a frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            else if (mstp_port->DataAvailable == TRUE)
            {
                /* Preamble2 */
                if (mstp_port->DataRegister == 0xFF)
                {
                    d_fr_printf("R_PREAMBLE Preamble2\n");
                    mstp_port->DataAvailable = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->RxIndex = 0;
                    mstp_port->HeaderCRC = 0xFF;
                    /* receive the remainder of the frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                }
                /* ignore RepeatedPreamble1 */
                else if (mstp_port->DataRegister == 0x55)
                {
                    d_fr_printf("R_PREAMBLE RepeatedPreamble1\n");
                    mstp_port->DataAvailable = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    /* wait for the second preamble octet. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_PREAMBLE;
                }
                /* NotPreamble */
                else
                {
                    d_fr_printf("R_PREAMBLE NotPreamble\n");
                    mstp_port->DataAvailable = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    /* wait for the start of a frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
                }
            }
            break;
            /* In the HEADER state, the node waits for the fixed message header. */
        case MSTP_RECEIVE_STATE_HEADER:
            /* Timeout */
            if (mstp_port->SilenceTimer > mstp_port->Tframe_abort)
            {
                fr_printf("R_HEADER Timeout\n");
                /* indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of a frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }

            /* Error */
            else if (mstp_port->ReceiveError == TRUE)
            {
                fr_printf("R_HEADER Error\n");
                mstp_port->ReceiveError = FALSE;
                //mstp_port->SilenceTimer = 0;
                mstp_port->EventCount++;
                /* indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of a frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            else if (mstp_port->DataAvailable == TRUE)
            {
                /* FrameType */
                if (mstp_port->RxIndex == 0)
                {
                    d_fr_printf("R_HEADER FrameType\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->FrameType = mstp_port->DataRegister;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex = 1;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                }
                /* Destination */
                else if (mstp_port->RxIndex == 1)
                {
                    d_fr_printf("R_HEADER Destination\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->DestinationAddress = mstp_port->DataRegister;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex = 2;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                }
                /* Source */
                else if (mstp_port->RxIndex == 2)
                {
                    d_fr_printf("R_HEADER Source\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->SourceAddress = mstp_port->DataRegister;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex = 3;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                    if (mstp_port->SourceAddress == mstp_port->This_Station) {
                        indicate_duplicated_mac();
                    }
                }
                /* Length1 */
                else if (mstp_port->RxIndex == 3)
                {
                    d_fr_printf("R_HEADER Length1\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->DataLength = mstp_port->DataRegister * 256;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex = 4;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                }
                /* Length2 */
                else if (mstp_port->RxIndex == 4)
                {
                    d_fr_printf("R_HEADER Length2\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->DataLength += mstp_port->DataRegister;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex = 5;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER;
                }
                /* HeaderCRC */
                else if (mstp_port->RxIndex == 5)
                {
                    d_fr_printf("R_HEADER HeaderCRC\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    mstp_port->HeaderCRC = CRC_Calc_Header(mstp_port->DataRegister, mstp_port->HeaderCRC);
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_HEADER_CRC;
                }
                /* not per MS/TP standard, but it is a case not covered */
                else
                {
                    fr_printf("R_HEADER case not covered\n");
                    mstp_port->ReceiveError = FALSE;
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->EventCount++;
                    /* indicate that an error has occurred during  */
                    /* the reception of a frame */
                    mstp_port->ReceivedInvalidFrame = TRUE;
                    /* wait for the start of a frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
                }
            }
            break;
            /* In the HEADER_CRC state, the node validates the CRC on the fixed message header. */
        case MSTP_RECEIVE_STATE_HEADER_CRC:
            /* BadCRC */
            if (mstp_port->HeaderCRC != 0x55)
            {
                fr_printf("R_HEADER_CRC BadCRC\n");
                /* indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of the next frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            else
            {
                /* we have received a good frame */
                mstp_port->looking_for_frame=0;
                /* use this to notify higher layers that we are not alone */
                /* even tho we do not know the mac of next master */
                indicate_sole_master(FALSE);
                /* have we found the correct baud rate */
                if (mstp_port->found_frames < AUTO_BOUD_BAUD_FOUND) mstp_port->found_frames++;

                if ((mstp_port->DestinationAddress == mstp_port->This_Station) ||
                     (mstp_port->DestinationAddress == MSTP_BROADCAST_ADDRESS))
                {
                    /* FrameTooLong */
                    if (mstp_port->DataLength > MAX_NPDU_SIZE)
                    {
                        fr_printf("R_HEADER_CRC FrameTooLong\n");
                        /* indicate that a frame with an illegal or  */
                        /* unacceptable data length has been received */
                        mstp_port->ReceivedInvalidFrame = TRUE;
                        /* wait for the start of the next frame. */
                        mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
                    }
                    /* NoData */
                    else if (mstp_port->DataLength == 0)
                    {
                        d_fr_printf("R_HEADER_CRC NoData\n");
                        /* indicate that a frame with no data has been received */
                        mstp_port->ReceivedValidFrame = TRUE;
                        /* wait for the start of the next frame. */
                        mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
                    }
                    /* Data */
                    else
                    {
                        d_fr_printf("R_HEADER_CRC Data\n");
                        /* Now it is too late for the upper layers to */
                        /* read from the data buffer (ignore error on sem==0) */
                        if (bac_sem_trywait(mstp_port->npdu_read_sem) == 0 ) {
                            fr_printf("missed read!!!!!!!!!!!!!!!!!\n");
                            /* Could not read the cell since we don't have a buffer */
                            mstp_port->ReceivedInvalidFrame = TRUE;
                            /* wait for the start of the next frame. */
                            mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;

                            /* save the error */
                            mstp_port->npdu_r_err |= READ_ERR_MISSED_FRAME;
                        }

                        d_fr_printf("R_HEADER_CRC Data\n");
                        mstp_port->RxIndex = 0;
                        mstp_port->DataCRC = 0xFFFF;
                        /* receive the data portion of the frame. */
                        mstp_port->receive_state = MSTP_RECEIVE_STATE_DATA;
                    }
                }
                /* NotForUs */
                else
                {
                    d_fr_printf("R_HEADER_CRC NotForUs\n");
                    /* wait for the start of the next frame. */
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
                }
            }
            break;
            /* In the DATA state, the node waits for the data portion of a frame. */
        case MSTP_RECEIVE_STATE_DATA:
            /* Timeout */
            if (mstp_port->SilenceTimer > mstp_port->Tframe_abort)
            {
                fr_printf("DATA Timeout\n");
                /* indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of the next frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            /* Error */
            else if (mstp_port->ReceiveError == TRUE)
            {
                fr_printf("DATA Error\n");
                mstp_port->ReceiveError = FALSE;
                //mstp_port->SilenceTimer = 0;
                /* indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of the next frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            else if (mstp_port->DataAvailable == TRUE)
            {
                /* DataOctet */
                if (mstp_port->RxIndex < mstp_port->DataLength)
                {
                    d_fr_printf("DATA DataOctet\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->DataCRC = CRC_Calc_Data(mstp_port->DataRegister, mstp_port->DataCRC);
                    mstp_port->InputBuffer[mstp_port->RxIndex] = mstp_port->DataRegister;
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex++;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_DATA;
                }
                /* CRC1 */
                else if (mstp_port->RxIndex == mstp_port->DataLength)
                {
                    d_fr_printf("DATA CRC1\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->DataCRC = CRC_Calc_Data(mstp_port->DataRegister, mstp_port->DataCRC);
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->RxIndex++;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_DATA;
                }
                /* CRC2 */
                else if (mstp_port->RxIndex == (mstp_port->DataLength + 1))
                {
                    d_fr_printf("DATA CRC2\n");
                    //mstp_port->SilenceTimer = 0;
                    mstp_port->DataCRC = CRC_Calc_Data(mstp_port->DataRegister, mstp_port->DataCRC);
                    mstp_port->DataAvailable = FALSE;
                    mstp_port->receive_state = MSTP_RECEIVE_STATE_DATA_CRC;
                }
            }
            break;
            /* In the DATA_CRC state, the node validates the CRC of the message data. */
        case MSTP_RECEIVE_STATE_DATA_CRC:
            /* GoodCRC */
            if (mstp_port->DataCRC == 0xF0B8)
            {
                /* we have received a good frame */
                mstp_port->looking_for_frame=0;

                d_fr_printf("DATA_CRC GoodCRC\n");
                /* indicate the complete reception of a valid frame */
                mstp_port->ReceivedValidFrame = TRUE;
                mstp_port->DataAvailable = FALSE;

                /* save the length, source address and FrameTypefor delayed processing */
                mstp_port->npdu_r_len=mstp_port->RxIndex-1;
                mstp_port->npdu_r_src_mac=mstp_port->SourceAddress;
                mstp_port->npdu_r_frame_type=mstp_port->FrameType;

                MSTP_Master_Node_State_Machine();
                /* wait for the start of the next frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            /* BadCRC */
            else
            {
                fr_printf("DATA_CRC BadCRC\n");
                /* to indicate that an error has occurred during the reception of a frame */
                mstp_port->ReceivedInvalidFrame = TRUE;
                /* wait for the start of the next frame. */
                mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            }
            break;
        default:
            /* shouldn't get here - but if we do... */
            mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
            break;
    }
    return;
}

/****************************************************************/
void MSTP_Master_Node_State_Machine(void)
{
    INT8U tmp_poll_station;

    /* The master state machine must be run to the end before reenterd */
    switch (mstp_port->master_state)
    {
        /* INITIALIZE state is entered in MSTP_Init and then goes to MSTP_MASTER_STATE_IDLE */
        /* In the IDLE state, the node waits for a frame. */
        case MSTP_MASTER_STATE_INITIALIZE:
        case MSTP_MASTER_STATE_IDLE:
            /* LostToken */
            if (mstp_port->SilenceTimer >= Tno_token)
            {
                ma_printf("M_IDLE LostToken\n");
                /* assume that the token has been lost */
                mstp_port->master_state = MSTP_MASTER_STATE_NO_TOKEN;
            }
            /* mstp_port->ReceivedInvalidFrame */
            else if (mstp_port->ReceivedInvalidFrame == TRUE)
            {
                indicate_frame_error();

                d_ma_printf("M_IDLE ReceivedInvalidFrame\n");
                /* invalid frame was received */
                mstp_port->ReceivedInvalidFrame = FALSE;
                /* wait for the next frame */
                mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            }
            else if (mstp_port->ReceivedValidFrame == TRUE)
            {
                /* ReceivedUnwantedFrame */
                if ((mstp_port->DestinationAddress != mstp_port->This_Station) &&
                     (mstp_port->DestinationAddress != MSTP_BROADCAST_ADDRESS))
                {
                    ma_printf("M_IDLE UnwantedFrame. != This_Station && != BROADCAST\n");
                    /* an unexpected or unwanted frame was received. */
                    mstp_port->ReceivedValidFrame = FALSE;
                    /* wait for the next frame */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                }
                /* DestinationAddress is equal to 255 (broadcast) and  */
                /* FrameType has a value of Token, BACnet Data Expecting Reply, Test_Request,  */
                /* or a proprietary type known to this node that expects a reply */
                /* (such frames may not be broadcast), or */
                else if ((mstp_port->DestinationAddress == MSTP_BROADCAST_ADDRESS) &&
                          ((mstp_port->FrameType == FRAME_TYPE_TOKEN) ||
                          (mstp_port->FrameType == FRAME_TYPE_BACNET_DATA_EXPECTING_REPLY) ||
                          (mstp_port->FrameType == FRAME_TYPE_TEST_REQUEST)))
                {
                    ma_printf("M_IDLE UnwantedFrame. B_CAST && (TOKEN || EXP_REPLY || TEST_REQ)\n");
                    /* an unexpected or unwanted frame was received. */
                    mstp_port->ReceivedValidFrame = FALSE;
                    /* wait for the next frame */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                }
                /* FrameType has a value that indicates a standard or proprietary type */
                /* that is not known to this node. */
                /* change this if you add a proprietary type */
                else if /*(*/(mstp_port->FrameType >= FRAME_TYPE_PROPRIETARY_MIN) /*&&*/
                    /*(FrameType <= FRAME_TYPE_PROPRIETARY_MAX))*/
                    /* unnecessary if FrameType is INT8U with max of 255 */
                {
                    ma_printf("M_IDLE UnwantedFrame. >= FRAME_TYPE_PROPRIETARY_MIN\n");
                    /* an unexpected or unwanted frame was received. */
                    mstp_port->ReceivedValidFrame = FALSE;
                    /* wait for the next frame */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                }
                /* ReceivedToken */
                else if ((mstp_port->DestinationAddress == mstp_port->This_Station) &&
                          (mstp_port->FrameType == FRAME_TYPE_TOKEN))
                {
                    d_ma_printf("M_IDLE ReceivedToken\n");
                    mstp_port->ReceivedValidFrame = FALSE;
                    mstp_port->FrameCount = 0;
                    mstp_port->SoleMaster = FALSE;
                    indicate_sole_master(FALSE);
                    mstp_port->master_state = MSTP_MASTER_STATE_USE_TOKEN;
                }
                /* ReceivedPFM */
                else if ((mstp_port->DestinationAddress == mstp_port->This_Station) &&
                          (mstp_port->FrameType == FRAME_TYPE_POLL_FOR_MASTER))
                {
                    ma_printf("M_IDLE ReceivedPFM\n");

                    if (MSTP_Send_Controll_Frame(FRAME_TYPE_REPLY_TO_POLL_FOR_MASTER, mstp_port->SourceAddress) == FALSE)
                    {
                        /* re enter state later */
                        return;
                    }

                    mstp_port->ReceivedValidFrame = FALSE;
                    /* wait for the next frame */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                }
                /* ReceivedDataNoReply */
                /* or a proprietary type known to this node that does not expect a reply */
                else if (((mstp_port->DestinationAddress == mstp_port->This_Station) ||
                           (mstp_port->DestinationAddress == MSTP_BROADCAST_ADDRESS)) &&
                           ((mstp_port->FrameType == FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY) ||
                           /*  (mstp_port->FrameType == FRAME_TYPE_PROPRIETARY_0) || */
                           (mstp_port->FrameType == FRAME_TYPE_TEST_RESPONSE)))
                {
                    d_ma_printf("M_IDLE ReceivedDataNoReply\n");

                    mstp_port->ReceivedValidFrame = FALSE;
                    /* wait for the next frame */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;

                    /* Let the read proces read the data buffer */
                    bac_sem_post(mstp_port->npdu_read_sem);

                }
                /* ReceivedDataNeedingReply */
                /* or a proprietary type known to this node that expects a reply */
                else if (mstp_port->DestinationAddress == mstp_port->This_Station)
                {
                    if (mstp_port->FrameType == FRAME_TYPE_BACNET_DATA_EXPECTING_REPLY)
                        /* || (mstp_port->FrameType == FRAME_TYPE_PROPRIETARY) */
                    {
                        d_ma_printf("M_IDLE ReceivedDataNeedingReply (DATA_EXPECTING_REPLY)\n");
                        mstp_port->ReplyPostponedTimer = 0;
                        /* indicate successful reception to the higher layers  */
                        mstp_port->ReceivedValidFrame = FALSE;
                        mstp_port->master_state = MSTP_MASTER_STATE_ANSWER_DATA_REQUEST;

                        /* Let the read proces read the data buffer */
                        bac_sem_post(mstp_port->npdu_read_sem);

                    }
                    if (mstp_port->FrameType == FRAME_TYPE_TEST_REQUEST)
                    {
                        d_ma_printf("M_IDLE ReceivedDataNeedingReply (TEST_REQUEST)\n");
                        mstp_port->ReplyPostponedTimer = 0;
                        /* The reply to a TestRequest is done here in the MSTP stack */
                        mstp_port->ReceivedValidFrame = FALSE;
                        mstp_port->master_state = MSTP_MASTER_STATE_ANSWER_DATA_REQUEST;

                        /* Let the read proces read the data buffer */
                        bac_sem_post(mstp_port->npdu_read_sem);
                    }
                }
            }
            break;
        case MSTP_MASTER_STATE_USE_TOKEN:
            /* In the USE_TOKEN state, the node is allowed to send one or  */
            /* more data frames. These may be BACnet Data frames or */
            /* proprietary frames. */

            /* SendNoWait */
            /* If there is a frame awaiting transmission that */
            /* is of type Test_Response, BACnet Data Not Expecting Reply,  */
            /* or a proprietary type that does not expect a reply, */
            if ((mstp_port->SendType == MSTP_SEND_NO_WAIT) ||
                 (mstp_port->SendType == MSTP_SEND_REPLY) ||
                 (mstp_port->SendType == MSTP_SEND_TEST_RESPONSE))
            {
                d_ma_printf("USE_TOKEN SendNoWait\n");
                /* transmit the data frame */
                if (MSTP_Send_Data_Frame(mstp_port->O_Buffer, mstp_port->O_Buffer_Len) == FALSE)
                {
                    /* Re enter same state later to send frame */
                    return;
                }

                mstp_port->SendType = MSTP_SEND_NOTHING;
                mstp_port->FrameCount++;
                mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;
                /* Now the writing proces may continue */
                bac_sem_post(mstp_port->npdu_write_sem);
            }
            /* SendAndWait */
            /* If there is a frame awaiting transmission that is of  */
            /* type Test_Request, BACnet Data Expecting Reply, or  */
            /* a proprietary type that expects a reply, */
            else if (mstp_port->SendType == MSTP_SEND_AND_WAIT)
            {
                d_ma_printf("USE_TOKEN SendAndWait\n");
                /* transmit the data frame */
                if (MSTP_Send_Data_Frame(mstp_port->O_Buffer, mstp_port->O_Buffer_Len) == FALSE)
                {
                    /* Re enter same state later to send frame */
                    return;
                }
                mstp_port->SendType = MSTP_SEND_NOTHING;
                mstp_port->FrameCount++;
                mstp_port->master_state = MSTP_MASTER_STATE_WAIT_FOR_REPLY;
                /* In the WAIT_FOR_REPLY state, the node waits for  */
                /* a reply from another node. */

                /* Now the writing proces may continue */
                bac_sem_post(mstp_port->npdu_write_sem);
            }

            /* TestRequest */
            else if (mstp_port->SendType == MSTP_SEND_TEST_REQUEST)
            {
                d_ma_printf("USE_TOKEN TestRequest\n");
                /* transmit the data frame */
                if (MSTP_Send_Data_Frame(mstp_port->O_Buffer, mstp_port->O_Buffer_Len) == FALSE)
                {
                    /* Re enter same state later to send frame */
                    return;
                }
                mstp_port->SendType = MSTP_SEND_NOTHING;
                mstp_port->FrameCount++;
                mstp_port->master_state = MSTP_MASTER_STATE_WAIT_FOR_REPLY;
                /* In the WAIT_FOR_REPLY state, the node waits for  */
                /* a reply from another node. */

                /* Now the writing proces may continue */
                bac_sem_post(mstp_port->npdu_write_sem);
            }

            /* NothingToSend */
            else
            {
                d_ma_printf("USE_TOKEN NothingToSend\n");
                mstp_port->FrameCount = MAX_INFO_FRAMES;
                mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;
            }
            break;
        case MSTP_MASTER_STATE_WAIT_FOR_REPLY:
            /* ReplyTimeout */
            if (mstp_port->SilenceTimer >= Treply_timeout)
            {
                ma_printf("WAIT_FOR_REPLY ReplyTimeout\n");
                /* assume that the request has failed */

                mstp_port->FrameCount = MAX_INFO_FRAMES;
                mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;
                /* Any retry of the data frame shall await the next entry */
                /* to the USE_TOKEN state. (Because of the length of the timeout,  */
                /* this transition will cause the token to be passed regardless */
                /* of the initial value of FrameCount.) */

                /* Let the read proces read the data buffer -> report error */
                mstp_port->npdu_r_err |= READ_ERR_REPL_TMO;
                bac_sem_post(mstp_port->npdu_read_sem);
            }
            /* InvalidFrame */
            else if (mstp_port->ReceivedInvalidFrame == TRUE)
            {
                /* (mstp_port->SilenceTimer < Treply_timeout) */
                ma_printf("WAIT_FOR_REPLY InvalidFrame\n");
                /* error in frame reception */
                mstp_port->ReceivedInvalidFrame = FALSE;
                mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;

                /* What happends here to a read ? If it was this frame we were waiting for */
                /* we will receive a timeout and the "user" will be notifyed.*/
            }
            /* ReceivedReply */
            /* or a proprietary type that indicates a reply */
            else if (mstp_port->ReceivedValidFrame == TRUE)
            {
                if (mstp_port->DestinationAddress == mstp_port->This_Station)
                {
                    /*ReceivedReply*/
                    if (mstp_port->FrameType == FRAME_TYPE_TEST_RESPONSE)
                    {
                        /* Indicate successful reception to the higher layers */
                        ma_printf("WAIT_FOR_REPLY TestResponce\n");

                        mstp_port->ReceivedValidFrame = FALSE;
                        mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;
                        /* Let the read proces read the data buffer */
                        bac_sem_post(mstp_port->npdu_read_sem);

                    }
                    /*ReceivedReply*/
                    else if (mstp_port->FrameType == FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY)
                    {
                        /* Indicate successful reception to the higher layers */
                        d_ma_printf("WAIT_FOR_REPLY ReceivedReply\n");

                        mstp_port->ReceivedValidFrame = FALSE;
                        mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;

                        /* Let the read proces read the data buffer */
                        bac_sem_post(mstp_port->npdu_read_sem);

                    }
                    /* ReceivedPostpone */
                    else if (mstp_port->FrameType == FRAME_TYPE_REPLY_POSTPONED)
                    {
                        /* (mstp_port->SilenceTimer < Treply_timeout) */
                        d_ma_printf("WAIT_FOR_REPLY ReceivedPostpone\n");
                        /* Then the reply to the message has been postponed until a later time. */

                        mstp_port->ReceivedValidFrame = FALSE;
                        mstp_port->master_state = MSTP_MASTER_STATE_DONE_WITH_TOKEN;

                        /* Let the read proces read the data buffer -> report error */
                        mstp_port->npdu_r_frame_type = FRAME_TYPE_REPLY_POSTPONED;
                        bac_sem_post(mstp_port->npdu_read_sem);
                    }
                    /* ReceivedUnexpectedFrame */
                    else {
                        /* !=FRAME_TYPE_TEST_RESPONSE, */
                        /* !=FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY */
                        /* !=FRAME_TYPE_REPLY_POSTPONED */
                        ma_printf("WAIT_FOR_REPLY ReceivedUnexpectedFrame 1\n");
                        one_p_printf("FRAME_TYPE==%d\n",mstp_port->FrameType);
                        /* an unexpected frame was received */
                        /* This may indicate the presence of multiple tokens.  */
                        mstp_port->ReceivedValidFrame = FALSE;
                        /* Synchronize with the network. */
                        /* This action drops the token.       */
                        mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                    }
                }
                /* ReceivedUnexpectedFrame */
                else /*the expected reply should not be broadcast)  */
                {
                    /* (mstp_port->SilenceTimer < Treply_timeout) */
                    ma_printf("WAIT_FOR_REPLY ReceivedUnexpectedFrame (reply should not be broadcast)\n");
                    one_p_printf("frame type   = %d\n",mstp_port->FrameType);
                    /* an unexpected frame was received */
                    /* This may indicate the presence of multiple tokens.  */
                    mstp_port->ReceivedValidFrame = FALSE;
                    /* Synchronize with the network. */
                    /* This action drops the token.       */
                    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
                }
            }
            break;
        case MSTP_MASTER_STATE_DONE_WITH_TOKEN:
            d_one_p_printf("M_DONE_WT FrameCount=%d\n", mstp_port->FrameCount);
            d_one_p_printf("M_DONE_WT TokenCount=%d\n", mstp_port->TokenCount);
            /* The DONE_WITH_TOKEN state either sends another data frame,  */
            /* passes the token, or initiates a Poll For Master cycle. */
            /* SendAnotherFrame */
            if (mstp_port->FrameCount < MAX_INFO_FRAMES)
            {
                ma_printf("M_DONE_WT SendAnotherFrame\n");
                /* then this node may send another information frame  */
                /* before passing the token.  */
                mstp_port->master_state = MSTP_MASTER_STATE_USE_TOKEN;
            }
            /* mstp_port->SoleMaster */
            else if ((mstp_port->FrameCount >= MAX_INFO_FRAMES) &&
                      (mstp_port->TokenCount < Npoll) &&
                      (mstp_port->SoleMaster == TRUE))
            {
                d_ma_printf("M_DONE_WT SoleMaster\n");
                /* there are no other known master nodes to */
                /* which the token may be sent (TRUE master-slave operation).  */
                mstp_port->FrameCount = 0;
                mstp_port->TokenCount++;
                mstp_port->master_state = MSTP_MASTER_STATE_USE_TOKEN;
            }
            /* SendToken */
            else if (((mstp_port->FrameCount >= MAX_INFO_FRAMES) &&
                       (mstp_port->TokenCount < Npoll) &&
                       (mstp_port->SoleMaster == FALSE)) ||
                       /* The comparison of NS and TS+1 eliminates the Poll For Master  */
                       /* if there are no addresses between TS and NS, since there is no  */
                       /* address at which a new master node may be found in that case. */
                       (mstp_port->Next_Station == (INT8U)((mstp_port->This_Station +1) % (MSTP_MAX_MASTER + 1))))
            {
                d_ma_printf("M_DONE_WT SendToken\n");
                /* transmit a Token frame to NS */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_TOKEN, mstp_port->Next_Station) == FALSE)
                {
                    /* Re enter same state later to send frame */
                    return;
                }
                mstp_port->TokenCount++;
                mstp_port->RetryCount = 0;
                mstp_port->EventCount = 0;
                mstp_port->master_state = MSTP_MASTER_STATE_PASS_TOKEN;
            }
            /* SendMaintenancePFM */
            else if ((mstp_port->FrameCount >= MAX_INFO_FRAMES) &&
                      (mstp_port->TokenCount >= Npoll) &&
                      ((INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1)) != mstp_port->Next_Station))
            {
                d_ma_printf("M_DONE_WT SendMaintenancePFM\n");

                tmp_poll_station = (INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1));
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_POLL_FOR_MASTER, tmp_poll_station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->Poll_Station = tmp_poll_station;
                mstp_port->RetryCount = 0;
                mstp_port->master_state = MSTP_MASTER_STATE_POLL_FOR_MASTER;
            }
            /* ResetMaintenancePFM */
            else if ((mstp_port->FrameCount >= MAX_INFO_FRAMES) &&
                      (mstp_port->TokenCount >= Npoll) &&
                      ((INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1)) == mstp_port->Next_Station) &&
                      (mstp_port->SoleMaster == FALSE))
            {
                d_ma_printf("M_DONE_WT ResetMaintenancePFM\n");
                /* transmit a Token frame to NS */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_TOKEN, mstp_port->Next_Station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->Poll_Station = mstp_port->This_Station;
                mstp_port->RetryCount = 0;
                mstp_port->TokenCount = 0;
                mstp_port->EventCount = 0;
                mstp_port->master_state = MSTP_MASTER_STATE_PASS_TOKEN;

                if (mstp_port->notify_on_reset_pfm==TRUE)
                {
                    /* This is a hack to send an "I am" when first connected to the network */
                    mstp_port->notify_on_reset_pfm=FALSE;
                    mstp_port->npdu_r_err |= READ_RESTART_PFM;
                    bac_sem_post(mstp_port->npdu_read_sem);
                }
            }
            /* SoleMasterRestartMaintenancePFM */
            else if ((mstp_port->FrameCount >= MAX_INFO_FRAMES) &&
                      (mstp_port->TokenCount >= Npoll) &&
                      ((INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1)) == mstp_port->Next_Station) &&
                      (mstp_port->SoleMaster == TRUE))
            {
                d_ma_printf("M_DONE_WT SoleMasterRestartMaintenancePFM\n");
                tmp_poll_station = (INT8U)((mstp_port->Next_Station +1) % (MSTP_MAX_MASTER + 1));
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_POLL_FOR_MASTER, tmp_poll_station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                /* no known successor node */
                mstp_port->Poll_Station = tmp_poll_station;
                mstp_port->Next_Station = mstp_port->This_Station;
                mstp_port->RetryCount = 0;
                /* NOTE: I think this is a bug in the standard.          */
                /* If I set TokenCount here to 0 I will have to wait for */
                /* the state mashine to increment TokenCount to Npoll    */
                /* before the next "poll for master frame is sent.       */
                /* Or then the BLT test specification (2.2.7) is wrong!  */
                /* (added by Kåre Särs 2005)                             */
                /* mstp_port->TokenCount = 0; */
                mstp_port->TokenCount = Npoll;
                mstp_port->EventCount = 0;
                /* find a new successor to TSi */
                mstp_port->master_state = MSTP_MASTER_STATE_POLL_FOR_MASTER;
                indicate_sole_master(TRUE);
            }
            break;
        case MSTP_MASTER_STATE_PASS_TOKEN:
            /* The PASS_TOKEN state listens for a successor to begin using */
            /* the token that this node has just attempted to pass. */
            /* SawTokenUser */
            if ((mstp_port->SilenceTimer < Tusage_timeout) &&
                 (mstp_port->EventCount > Nmin_octets))
            {
                d_ma_printf("M_PASS_TOKEN SawTokenUser\n");
                /* Assume that a frame has been sent by the new token user.  */
                /* Enter the IDLE state to process the frame. */
                mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            }
            /* RetrySendToken */
            else if ((mstp_port->SilenceTimer >= Tusage_timeout) &&
                      (mstp_port->RetryCount < Nretry_token))
            {
                ma_printf("M_PASS_TOKEN RetrySendToken\n");
                /* Transmit a Token frame to NS */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_TOKEN, mstp_port->Next_Station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->RetryCount++;
                mstp_port->EventCount = 0;
                /* re-enter the current state to listen for NS  */
                /* to begin using the token. */
            }
            /* FindNewSuccessor */
            else if ((mstp_port->SilenceTimer >= Tusage_timeout) && (mstp_port->RetryCount >= Nretry_token))
            {
                ma_printf("M_PASS_TOKEN FindNewSuccessor\n");
                /* Assume that NS has failed.  */
                tmp_poll_station = (INT8U)((mstp_port->Next_Station + 1) % (MSTP_MAX_MASTER + 1));
                /* Transmit a Poll For Master frame to PS. */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_POLL_FOR_MASTER, tmp_poll_station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                /* no known successor node */
                mstp_port->Poll_Station = tmp_poll_station;
                mstp_port->Next_Station = mstp_port->This_Station;
                mstp_port->RetryCount = 0;
                mstp_port->TokenCount = 0;
                mstp_port->EventCount = 0;
                /* find a new successor to TS */
                mstp_port->master_state = MSTP_MASTER_STATE_POLL_FOR_MASTER;
            }
            else {
                /*ma_printf("R_PASS_TOKEN Don't care\n");*/
            }
            break;
        case MSTP_MASTER_STATE_NO_TOKEN:
            /* The NO_TOKEN state is entered if mstp_port->SilenceTimer becomes greater  */
            /* than Tno_token, indicating that there has been no network activity */
            /* for that period of time. The timeout is continued to determine  */
            /* whether or not this node may create a token. */

            /* SawFrame */
            if ((mstp_port->SilenceTimer < (Tno_token + (Tslot * mstp_port->This_Station))) &&
                 (mstp_port->EventCount > Nmin_octets))
            {
                one_p_printf("M_NO_TOKEN SawFrame (SilenceTimer=%d)\n", mstp_port->SilenceTimer);
                /* Some other node exists at a lower address.  */
                /* Enter the IDLE state to receive and process the incoming frame. */
                mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            }
            /* GenerateToken */
            else if ((mstp_port->SilenceTimer >= (Tno_token + (Tslot * mstp_port->This_Station))) &&
                      (mstp_port->SilenceTimer < (Tno_token + (Tslot * (mstp_port->This_Station + 1)))))
            {
                one_p_printf("M_NO_TOKEN GenerateToken (SilenceTimer=%d)\n", mstp_port->SilenceTimer);
                /* Assume that this node is the lowest numerical address  */
                /* on the network and is empowered to create a token.  */
                tmp_poll_station = (INT8U)((mstp_port->This_Station + 1) % (MSTP_MAX_MASTER + 1));
                /* Transmit a Poll For Master frame to PS. */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_POLL_FOR_MASTER, tmp_poll_station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                /* indicate that the next station is unknown */
                mstp_port->Poll_Station = tmp_poll_station;
                mstp_port->Next_Station = mstp_port->This_Station;
                mstp_port->RetryCount = 0;
                mstp_port->TokenCount = 0;
                mstp_port->EventCount = 0;
                /* enter the POLL_FOR_MASTER state to find a new successor to TS. */
                mstp_port->master_state = MSTP_MASTER_STATE_POLL_FOR_MASTER;
            }
            else {
                /*ma_printf("R_Don't care\n");*/
            }
            break;
        case MSTP_MASTER_STATE_POLL_FOR_MASTER:
            /* In the POLL_FOR_MASTER state, the node listens for a reply to */
            /* a previously sent Poll For Master frame in order to find  */
            /* a successor node. */
            /* ReceivedReplyToPFM */
            if ((mstp_port->ReceivedValidFrame == TRUE) &&
                 (mstp_port->DestinationAddress == mstp_port->This_Station) &&
                 (mstp_port->FrameType == FRAME_TYPE_REPLY_TO_POLL_FOR_MASTER))
            {
                ma_printf("M_PFM_ReceivedReplyToPFM\n");

                mstp_port->Next_Station = mstp_port->SourceAddress;
                /* Transmit a Token frame to NS */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_TOKEN, mstp_port->Next_Station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->SoleMaster = FALSE;
                indicate_sole_master(FALSE);
                mstp_port->EventCount = 0;
                mstp_port->Poll_Station = mstp_port->This_Station;
                mstp_port->TokenCount = 0;
                mstp_port->RetryCount = 0;
                mstp_port->ReceivedValidFrame = FALSE;
                mstp_port->master_state = MSTP_MASTER_STATE_PASS_TOKEN;
            }
            /* ReceivedUnexpectedFrame */
            else if ((mstp_port->ReceivedValidFrame == TRUE) &&
                      ((mstp_port->DestinationAddress != mstp_port->This_Station) ||
                      (mstp_port->FrameType != FRAME_TYPE_REPLY_TO_POLL_FOR_MASTER)))
            {
                ma_printf("M_PFM_ReceivedUnexpectedFrame\n");
                /* An unexpected frame was received.  */
                /* This may indicate the presence of multiple tokens. */
                mstp_port->ReceivedValidFrame = FALSE;
                /* enter the IDLE state to synchronize with the network.  */
                /* This action drops the token. */
                mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            }
            /* SoleMaster */
            else if ((mstp_port->SoleMaster == TRUE) &&
                      ((mstp_port->SilenceTimer >= Tusage_timeout) ||
                      (mstp_port->ReceivedInvalidFrame == TRUE)))
            {
                d_ma_printf("M_PFM_SoleMaster\n");
                /* There was no valid reply to the periodic poll  */
                /* by the sole known master for other masters. */
                mstp_port->FrameCount = 0;
                mstp_port->ReceivedInvalidFrame = FALSE;
                mstp_port->master_state = MSTP_MASTER_STATE_USE_TOKEN;
            }
            /* DoneWithPFM */
            else if ((mstp_port->SoleMaster == FALSE) &&
                      (mstp_port->Next_Station != mstp_port->This_Station) &&
                      ((mstp_port->SilenceTimer >= Tusage_timeout) ||
                      (mstp_port->ReceivedInvalidFrame == TRUE)))
            {
                d_ma_printf("M_PFM_DoneWithPFM\n");
                /* There was no valid reply to the maintenance poll for a master at address PS.  */
                /* transmit a Token frame to NS */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_TOKEN, mstp_port->Next_Station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->EventCount = 0;
                mstp_port->RetryCount = 0;
                mstp_port->ReceivedInvalidFrame = FALSE;
                mstp_port->master_state = MSTP_MASTER_STATE_PASS_TOKEN;
            }
            /* SendNextPFM */
            else if ((mstp_port->SoleMaster == FALSE) &&
                      (mstp_port->Next_Station == mstp_port->This_Station) && /* no known successor node */
                      ((INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1)) != mstp_port->This_Station) &&
                      ((mstp_port->SilenceTimer >= Tusage_timeout) || (mstp_port->ReceivedInvalidFrame == TRUE)))
            {
                tmp_poll_station = (INT8U)((mstp_port->Poll_Station + 1) % (MSTP_MAX_MASTER + 1));
                d_one_p_printf("M_PFM SendNextPFM SilenceTimer = %d\n", mstp_port->SilenceTimer);
                /* Transmit a Poll For Master frame to PS. */
                if (MSTP_Send_Controll_Frame(FRAME_TYPE_POLL_FOR_MASTER, tmp_poll_station) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                mstp_port->Poll_Station = tmp_poll_station;
                d_two_p_printf("M_PFM SendNextPFM from %d to %d\n", mstp_port->This_Station, mstp_port->Poll_Station);
                mstp_port->RetryCount = 0;
                mstp_port->ReceivedInvalidFrame = FALSE;
                /* Re-enter the current state. */
            }
            /* DeclareSoleMaster */
            else if ((mstp_port->SoleMaster == FALSE) &&
                      (mstp_port->Next_Station == mstp_port->This_Station) && /* no known successor node */
                      ((INT8U)((mstp_port->Poll_Station + 1) %
                      (MSTP_MAX_MASTER + 1)) == mstp_port->This_Station) &&
                      ((mstp_port->SilenceTimer >= Tusage_timeout) ||
                      (mstp_port->ReceivedInvalidFrame == TRUE)))
            {
                ma_printf("M_PFM DeclareSoleMaster\n");
                /* to indicate that this station is the only master */
                mstp_port->SoleMaster = TRUE;
                indicate_sole_master(TRUE);
                mstp_port->FrameCount = 0;
                /* added by Kåre Särs to allow immediate starting of the PFM cycle */
                mstp_port->TokenCount = Npoll;
                mstp_port->ReceivedInvalidFrame = FALSE;
                mstp_port->master_state = MSTP_MASTER_STATE_USE_TOKEN;
            }
            else if (mstp_port->ReceivedInvalidFrame == TRUE) {
                ma_printf("M_what is this? 1\n");
            }
            else if ((mstp_port->SilenceTimer >= Tusage_timeout)) {
                ma_printf("M_what is this? 2\n");
            }
            else {
                /*ma_printf("R_don't care\n");*/
            }
            break;
        case MSTP_MASTER_STATE_ANSWER_DATA_REQUEST:
            /* The ANSWER_DATA_REQUEST state is entered when a  */
            /* BACnet Data Expecting Reply, a Test_Request, or  */
            /* a proprietary frame that expects a reply is received. */
            if (mstp_port->ReplyPostponedTimer <= Treply_delay)
            {
                /* Reply */
                /* If a reply is available from the higher layers  */
                /* within Treply_delay after the reception of the  */
                /* final octet of the requesting frame  */
                /* (the mechanism used to determine this is a local matter), */
                /* then call MSTP_Send_Data_Frame to transmit the reply frame  */
                /* and enter the IDLE state to wait for the next frame. */
                /* ("mstp_port->FrameType" can be used since no other device */
                /* may send anything untill we respond or timeout */
                if (mstp_port->FrameType == FRAME_TYPE_BACNET_DATA_EXPECTING_REPLY)
                {
                    if (mstp_port->SendType == MSTP_SEND_REPLY)
                    {
                        d_ma_printf("ANSWER_DATA_REQUEST Reply\n");

                        if (MSTP_Send_Data_Frame(mstp_port->O_Buffer, mstp_port->O_Buffer_Len) == FALSE)
                        {
                            /* Re enter same state later to send frame */
                            return;
                        }

                        mstp_port->SendType = MSTP_SEND_NOTHING;
                        mstp_port->master_state = MSTP_MASTER_STATE_IDLE;

                        /* Now the write proces may continue */
                        bac_sem_post(mstp_port->npdu_write_sem);

                    }
                    else if (mstp_port->SendType != MSTP_SEND_NOTHING) {
                        /* We need the REPLY NOW! */
                        /* set the type to error (indication to user) and release the write. */
                        /* What ever was in the queue is discarded and we wait for the reply */
                        mstp_port->SendType = MSTP_SEND_ERROR;
                        bac_sem_post(mstp_port->npdu_write_sem);
                        /* renter the same state */
                    }
                }

                /* TestRequest */
                /* If a receiving node can successfully receive and return  */
                /* the information field, it shall do so. If it cannot receive */
                /* and return the entire information field but can detect  */
                /* the reception of a valid Test_Request frame  */
                /* (for example, by computing the CRC on octets as  */
                /* they are received), then the receiving node shall discard  */
                /* the information field and return a Test_Response containing  */
                /* no information field. If the receiving node cannot detect  */
                /* the valid reception of frames with overlength information fields,  */
                /* then no response shall be returned. */
                else if (mstp_port->FrameType == FRAME_TYPE_TEST_REQUEST)
                {
                    if (mstp_port->SendType == MSTP_SEND_TEST_RESPONSE)
                    {
                        d_ma_printf("ANSWER_DATA_REQUEST TestResponse\n");

                        if (MSTP_Send_Data_Frame(mstp_port->O_Buffer, mstp_port->O_Buffer_Len) == FALSE)
                        {
                            /* Re enter same state later to send frame */
                            return;
                        }

                        mstp_port->SendType = MSTP_SEND_NOTHING;
                        mstp_port->master_state = MSTP_MASTER_STATE_IDLE;

                        /* Now the write proces may continue */
                        bac_sem_post(mstp_port->npdu_write_sem);

                    }
                    else if (mstp_port->SendType != MSTP_SEND_NOTHING) {
                        /* We need the RESPONCE NOW! */
                        /* set the type to error (indication to user) and release the write. */
                        /* What ever was in the queue is discarded and we wait for the reply */
                        mstp_port->SendType = MSTP_SEND_ERROR;
                        bac_sem_post(mstp_port->npdu_write_sem);
                        /* renter the same state */
                    }

                }
            }

            /* */
            /* DeferredReply */
            /* If no reply will be available from the higher layers */
            /* within Treply_delay after the reception of the */
            /* final octet of the requesting frame (the mechanism */
            /* used to determine this is a local matter), */
            /* then an immediate reply is not possible. */
            /* Any reply shall wait until this node receives the token. */
            /* Call MSTP_Send_Controll_Frame to transmit a Reply Postponed frame, */
            /* and enter the IDLE state. */

            else
            {
                d_ma_printf("ANSWER_DATA_REQUEST DeferredReply\n");

                if (MSTP_Send_Controll_Frame(FRAME_TYPE_REPLY_POSTPONED,
                    mstp_port->SourceAddress) == FALSE)
                {
                    /* Re enter same state later */
                    return;
                }
                /* signal uper layers about postponing ? */
                mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            }
            break;
        default:
            one_p_printf("master_state==%d\n", mstp_port->master_state);
            mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
            break;
    }
    return;
}


/****************************************************************/
BOOLEAN MSTP_Init(struct mstp_port_struct_t *port, INT8U this_station_mac)
{
    mstp_port = port;

    if (mstp_port==NULL) return FALSE;

    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
    mstp_port->ReceiveError = FALSE;
    mstp_port->DataAvailable = FALSE;
    mstp_port->DataRegister = 0;
    mstp_port->HeaderCRC = 0;
    mstp_port->DataCRC = 0;
    mstp_port->DataLength = 0;
    mstp_port->DestinationAddress = 0;
    mstp_port->EventCount = 0;
    mstp_port->FrameType = FRAME_TYPE_TOKEN;
    mstp_port->FrameCount = 0;
    mstp_port->RxIndex = 0;
    mstp_port->TxIndex = 0;

    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
    mstp_port->Next_Station = this_station_mac;
    mstp_port->Poll_Station = this_station_mac;
    mstp_port->This_Station = this_station_mac;
    mstp_port->ReceivedInvalidFrame = FALSE;
    mstp_port->ReceivedValidFrame = FALSE;
    mstp_port->RetryCount = 0;
    mstp_port->SilenceTimer = 0;
    mstp_port->ReplyPostponedTimer = 0;
    mstp_port->SoleMaster = FALSE;
    mstp_port->SourceAddress = 0;
    /* cause a Poll For Master to be sent when this node first */
    /* receives the token */
    mstp_port->TokenCount = Npoll;

    mstp_port->SendType = MSTP_SEND_NOTHING;

    mstp_port->notify_on_reset_pfm = TRUE;
    mstp_port->looking_for_frame = 0;
    mstp_port->found_frames = 0;

    bac_sem_init(mstp_port->npdu_read_sem, 0);
    bac_sem_init(mstp_port->npdu_protect_read_sem, 1);
    bac_sem_init(mstp_port->npdu_write_sem, 0);
    bac_sem_init(mstp_port->npdu_protect_write_sem, 1);

    return TRUE;
}

/****************************************************************/
void MSTP_Re_Init(INT8U this_station_mac)
{

    mstp_port->receive_state = MSTP_RECEIVE_STATE_IDLE;
    mstp_port->ReceiveError = FALSE;
    mstp_port->DataAvailable = FALSE;
    mstp_port->DataRegister = 0;
    mstp_port->HeaderCRC = 0;
    mstp_port->DataCRC = 0;
    mstp_port->DataLength = 0;
    mstp_port->DestinationAddress = 0;
    mstp_port->EventCount = 0;
    mstp_port->FrameType = FRAME_TYPE_TOKEN;
    mstp_port->FrameCount = 0;
    mstp_port->RxIndex = 0;
    mstp_port->TxIndex = 0;

    mstp_port->master_state = MSTP_MASTER_STATE_IDLE;
    mstp_port->Next_Station = this_station_mac;
    mstp_port->Poll_Station = this_station_mac;
    mstp_port->This_Station = this_station_mac;
    mstp_port->ReceivedInvalidFrame = FALSE;
    mstp_port->ReceivedValidFrame = FALSE;
    mstp_port->RetryCount = 0;
    mstp_port->SilenceTimer = 0;
    mstp_port->ReplyPostponedTimer = 0;
    mstp_port->SoleMaster = FALSE;
    mstp_port->SourceAddress = 0;
    /* cause a Poll For Master to be sent when this node first */
    /* receives the token */
    mstp_port->TokenCount = Npoll;

    if (mstp_port->SendType != MSTP_SEND_NOTHING) {
        /* We have a task waiting to send stuff */
        /* release the task and send nothing */
        bac_sem_post(mstp_port->npdu_write_sem);

        mstp_port->SendType = MSTP_SEND_NOTHING;
    }

    mstp_port->looking_for_frame = 0;
    mstp_port->found_frames = 0;

    no_p_printf("Reinit mstp\n");

    return;
}

/************************************************************/
INT16U MSTP_read(INT8U *buf, INT16U len, INT8U *from_mac, INT16U *type)
{

    INT16U i, ret;

    bac_sem_wait(mstp_port->npdu_protect_read_sem);
    *type = 0;

    /* check the length of the buffer */
    if (len < MAX_NPDU_SIZE)
    {
        bac_sem_post(mstp_port->npdu_protect_read_sem);
        *type = READ_ERR_LEN;
        return 0;
    }

    /* wait for data to be ready */
    /* we must have a thread waiting otherwise we don't have buffer to save to :) */
    bac_sem_wait(mstp_port->npdu_read_sem);

    /* copy source address */
    *from_mac = mstp_port->npdu_r_src_mac;
    mstp_port->npdu_r_src_mac = 0xFF;

    /* copy the data */
    for (i=0; i<mstp_port->npdu_r_len; i++)
    {
        buf[i]=mstp_port->InputBuffer[i];
    }

    *type = mstp_port->npdu_r_frame_type | mstp_port->npdu_r_err;

    ret = mstp_port->npdu_r_len;

    /* reset old values */
    mstp_port->npdu_r_err = 0;
    mstp_port->npdu_r_len = 0;
    mstp_port->npdu_r_frame_type = 0;

    /* let next instance continue */
    bac_sem_post(mstp_port->npdu_protect_read_sem);

    /* return the number of bytes received */
    return ret;
}


/************************************************************/
INT16U MSTP_write(INT8U *data, INT16U data_len, INT8U to_mac, MSTP_WRITE_TYPE type)
{
    INT8U frame_type;

    /* If one instance of this function is waiting, wait here */
    bac_sem_wait(mstp_port->npdu_protect_write_sem);

    /* we can not write more than the buffer allows */
    if (data_len > MAX_NPDU_SIZE) {
        bac_sem_post(mstp_port->npdu_protect_write_sem);
        return WRITE_ERR_LEN;
    }

    /* send mode */
    switch (type)
    {
        case MSTP_SEND_NO_WAIT:
            frame_type = FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY;
            break;

        case MSTP_SEND_AND_WAIT:
            frame_type = FRAME_TYPE_BACNET_DATA_EXPECTING_REPLY;
            break;

        case MSTP_SEND_REPLY:
            frame_type = FRAME_TYPE_BACNET_DATA_NOT_EXPECTING_REPLY;
            break;

        case MSTP_SEND_TEST_REQUEST:
            frame_type = FRAME_TYPE_TEST_REQUEST;
            break;

        case MSTP_SEND_TEST_RESPONSE:
            frame_type = FRAME_TYPE_TEST_RESPONSE;
            break;

        default :
            bac_sem_post(mstp_port->npdu_protect_write_sem);
            return WRITE_ERR_WRONG_TYPE;
    }

    /* make MSTP frame */
    mstp_port->O_Buffer_Len = MSTP_Create_Frame(mstp_port->O_Buffer,
            MSTP_PDU_SIZE,
            frame_type,
            to_mac,
            data,
            data_len);

    mstp_port->SendType = (INT8U)type;

    /* wait for data to be sent */
    bac_sem_wait(mstp_port->npdu_write_sem);

    /* let next instance continue */
    bac_sem_post(mstp_port->npdu_protect_write_sem);

    /* if we are overwriten by some other task SendType is == MSTP_SEND_ERROR */
    if (mstp_port->SendType == MSTP_SEND_ERROR) {
        return WRITE_ERR_PREEMTIED;
    }
    return data_len;
}


